@@ -291,45 +291,45 @@ typedef struct
 /* ---------------------- vfat -------------------- */
 
 static const gchar *vfat_defaults[] = { "uid=", "gid=", "shortname=mixed", "utf8=1", "showexec", "flush", NULL };
-static const gchar *vfat_allow[] = { "flush", "utf8=", "shortname=", "umask=", "dmask=", "fmask=", "codepage=", "iocharset=", "usefree", "showexec", NULL };
-static const gchar *vfat_allow_uid_self[] = { "uid=", NULL };
-static const gchar *vfat_allow_gid_self[] = { "gid=", NULL };
+static const gchar *vfat_allow[] = { "flush", "utf8", "shortname", "umask", "dmask", "fmask", "codepage", "iocharset", "usefree", "showexec", NULL };
+static const gchar *vfat_allow_uid_self[] = { "uid", NULL };
+static const gchar *vfat_allow_gid_self[] = { "gid", NULL };
 
 /* ---------------------- ntfs -------------------- */
 /* this is assuming that ntfs-3g is used */
 
 static const gchar *ntfs_defaults[] = { "uid=", "gid=", NULL };
-static const gchar *ntfs_allow[] = { "umask=", "dmask=", "fmask=", "locale=", "norecover", "ignore_case", "windows_names", "compression", "nocompression", "big_writes", NULL };
-static const gchar *ntfs_allow_uid_self[] = { "uid=", NULL };
-static const gchar *ntfs_allow_gid_self[] = { "gid=", NULL };
+static const gchar *ntfs_allow[] = { "umask", "dmask", "fmask", "locale", "norecover", "ignore_case", "windows_names", "compression", "nocompression", "big_writes", NULL };
+static const gchar *ntfs_allow_uid_self[] = { "uid", NULL };
+static const gchar *ntfs_allow_gid_self[] = { "gid", NULL };
 
 /* ---------------------- iso9660 -------------------- */
 
 static const gchar *iso9660_defaults[] = { "uid=", "gid=", "iocharset=utf8", "mode=0400", "dmode=0500", NULL };
-static const gchar *iso9660_allow[] = { "norock", "nojoliet", "iocharset=", "mode=", "dmode=", NULL };
-static const gchar *iso9660_allow_uid_self[] = { "uid=", NULL };
-static const gchar *iso9660_allow_gid_self[] = { "gid=", NULL };
+static const gchar *iso9660_allow[] = { "norock", "nojoliet", "iocharset", "mode", "dmode", NULL };
+static const gchar *iso9660_allow_uid_self[] = { "uid", NULL };
+static const gchar *iso9660_allow_gid_self[] = { "gid", NULL };
 
 /* ---------------------- udf -------------------- */
 
 static const gchar *udf_defaults[] = { "uid=", "gid=", "iocharset=utf8", NULL };
-static const gchar *udf_allow[] = { "iocharset=", "umask=", NULL };
-static const gchar *udf_allow_uid_self[] = { "uid=", NULL };
-static const gchar *udf_allow_gid_self[] = { "gid=", NULL };
+static const gchar *udf_allow[] = { "iocharset", "umask", NULL };
+static const gchar *udf_allow_uid_self[] = { "uid", NULL };
+static const gchar *udf_allow_gid_self[] = { "gid", NULL };
 
 /* ---------------------- exfat -------------------- */
 
 static const gchar *exfat_defaults[] = { "uid=", "gid=", "iocharset=utf8", "namecase=0", "errors=remount-ro", NULL };
-static const gchar *exfat_allow[] = { "dmask=", "errors=", "fmask=", "iocharset=", "namecase=", "umask=", NULL };
-static const gchar *exfat_allow_uid_self[] = { "uid=", NULL };
-static const gchar *exfat_allow_gid_self[] = { "gid=", NULL };
+static const gchar *exfat_allow[] = { "dmask", "errors", "fmask", "iocharset", "namecase", "umask", NULL };
+static const gchar *exfat_allow_uid_self[] = { "uid", NULL };
+static const gchar *exfat_allow_gid_self[] = { "gid", NULL };
 
 /* ---------------------- hfs+ -------------------- */
 
 static const gchar *hfsplus_defaults[] = { "uid=", "gid=", "nls=utf8", NULL };
-static const gchar *hfsplus_allow[] = { "creator=", "type=", "umask=", "session=", "part=", "decompose", "nodecompose", "force", "nls=", NULL };
-static const gchar *hfsplus_allow_uid_self[] = { "uid=", NULL };
-static const gchar *hfsplus_allow_gid_self[] = { "gid=", NULL };
+static const gchar *hfsplus_allow[] = { "creator", "type", "umask", "session", "part", "decompose", "nodecompose", "force", "nls", NULL };
+static const gchar *hfsplus_allow_uid_self[] = { "uid", NULL };
+static const gchar *hfsplus_allow_gid_self[] = { "gid", NULL };
 
 /* ------------------------------------------------ */
 /* TODO: support context= */
@@ -360,12 +360,10 @@ find_mount_options_for_fs (const gchar *fstype)
     {
       fsmo = fs_mount_options + n;
       if (g_strcmp0 (fsmo->fstype, fstype) == 0)
-        goto out;
+        return fsmo;
     }
 
-  fsmo = NULL;
- out:
-  return fsmo;
+  return NULL;
 }
 
 static gid_t
@@ -442,62 +440,30 @@ is_uid_in_gid (uid_t uid,
 static gboolean
 is_mount_option_allowed (const FSMountOptions *fsmo,
                          const gchar          *option,
+                         const gchar          *value,
                          uid_t                 caller_uid)
 {
   int n;
   gchar *endp;
   uid_t uid;
   gid_t gid;
-  gboolean allowed;
-  const gchar *ep;
-  gsize ep_len;
-
-  allowed = FALSE;
 
   /* first run through the allowed mount options */
   if (fsmo != NULL)
     {
       for (n = 0; fsmo->allow != NULL && fsmo->allow[n] != NULL; n++)
         {
-          ep = strstr (fsmo->allow[n], "=");
-          if (ep != NULL && ep[1] == '\0')
+          if (strcmp (fsmo->allow[n], option) == 0)
             {
-              ep_len = ep - fsmo->allow[n] + 1;
-              if (strncmp (fsmo->allow[n], option, ep_len) == 0)
-                {
-                  allowed = TRUE;
-                  goto out;
-                }
-            }
-          else
-            {
-              if (strcmp (fsmo->allow[n], option) == 0)
-                {
-                  allowed = TRUE;
-                  goto out;
-                }
+              return TRUE;
             }
         }
     }
   for (n = 0; any_allow[n] != NULL; n++)
     {
-      ep = strstr (any_allow[n], "=");
-      if (ep != NULL && ep[1] == '\0')
+      if (strcmp (any_allow[n], option) == 0)
         {
-          ep_len = ep - any_allow[n] + 1;
-          if (strncmp (any_allow[n], option, ep_len) == 0)
-            {
-              allowed = TRUE;
-              goto out;
-            }
-        }
-      else
-        {
-          if (strcmp (any_allow[n], option) == 0)
-            {
-              allowed = TRUE;
-              goto out;
-            }
+          return TRUE;
         }
     }
 
@@ -509,15 +475,14 @@ is_mount_option_allowed (const FSMountOptions *fsmo,
       for (n = 0; fsmo->allow_uid_self != NULL && fsmo->allow_uid_self[n] != NULL; n++)
         {
           const gchar *r_mount_option = fsmo->allow_uid_self[n];
-          if (g_str_has_prefix (option, r_mount_option))
+          if (g_strcmp0 (option, r_mount_option) == 0)
             {
-              uid = strtol (option + strlen (r_mount_option), &endp, 10);
+              uid = strtol (value, &endp, 10);
               if (*endp != '\0')
                 continue;
               if (uid == caller_uid)
                 {
-                  allowed = TRUE;
-                  goto out;
+                  return TRUE;
                 }
             }
         }
@@ -530,36 +495,36 @@ is_mount_option_allowed (const FSMountOptions *fsmo,
       for (n = 0; fsmo->allow_gid_self != NULL && fsmo->allow_gid_self[n] != NULL; n++)
         {
           const gchar *r_mount_option = fsmo->allow_gid_self[n];
-          if (g_str_has_prefix (option, r_mount_option))
+          if (g_strcmp0 (option, r_mount_option) == 0)
             {
-              gid = strtol (option + strlen (r_mount_option), &endp, 10);
+              gid = strtol (value, &endp, 10);
               if (*endp != '\0')
                 continue;
               if (is_uid_in_gid (caller_uid, gid))
                 {
-                  allowed = TRUE;
-                  goto out;
+                  return TRUE;
                 }
             }
         }
     }
 
- out:
-  return allowed;
+  return FALSE;
 }
 
-static gchar **
+static GHashTable *
 prepend_default_mount_options (const FSMountOptions *fsmo,
                                uid_t                 caller_uid,
-                               GVariant             *given_options)
+                               GVariant             *given_options,
+                               gboolean              shared_fs)
 {
-  GPtrArray *options;
+  GHashTable *options;
   gint n;
   gchar *s;
   gid_t gid;
   const gchar *option_string;
 
-  options = g_ptr_array_new ();
+  options = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                   g_free, g_free);
   if (fsmo != NULL)
     {
       const gchar *const *defaults = fsmo->defaults;
@@ -567,25 +532,55 @@ prepend_default_mount_options (const FSMountOptions *fsmo,
       for (n = 0; defaults != NULL && defaults[n] != NULL; n++)
         {
           const gchar *option = defaults[n];
+          const gchar *eq = strchr (option, '=');
 
-          if (strcmp (option, "uid=") == 0)
+          if (eq != NULL)
             {
-              s = g_strdup_printf ("uid=%u", caller_uid);
-              g_ptr_array_add (options, s);
-            }
-          else if (strcmp (option, "gid=") == 0)
-            {
-              gid = find_primary_gid (caller_uid);
-              if (gid != (gid_t) - 1)
+              const gchar *value = eq + 1;
+              gsize opt_len = eq - option;
+              if (strncmp (option, "uid", opt_len) == 0)
                 {
-                  s = g_strdup_printf ("gid=%u", gid);
-                  g_ptr_array_add (options, s);
+                  s = g_strdup_printf ("%u", caller_uid);
+                  g_hash_table_insert (options, g_strdup ("uid"), s);
+                }
+              else if (strncmp (option, "gid", opt_len) == 0)
+                {
+                  gid = find_primary_gid (caller_uid);
+                  if (gid != (gid_t) - 1)
+                    {
+                      s = g_strdup_printf ("%u", gid);
+                      g_hash_table_insert (options, g_strdup ("gid"), s);
+                    }
+                }
+              else if (shared_fs && strncmp (option, "mode", opt_len) == 0)
+                {
+                  /* set different 'mode' and 'dmode' options for file systems mounted at shared
+                     location (otherwise they cannot be used by anybody else so mounting them at
+                     a shared location doesn't make much sense */
+                  gchar *shared_mode = g_strdup (value);
+
+                  /* give group and others the same permissions as to the owner
+                     without the 'write' permission, but at least 'read'
+                     (HINT: keep in mind that chars are ints in C and that
+                     digits are ordered naturally in the ASCII table) */
+                  shared_mode[2] = MAX(shared_mode[1] - 2, '4');
+                  shared_mode[3] = MAX(shared_mode[1] - 2, '4');
+                  g_hash_table_insert (options, g_strdup("mode"), shared_mode);
+                }
+              else if (shared_fs && strncmp (option, "dmode", opt_len) == 0)
+                {
+                  /* see right above */
+                  /* Does any other dmode than 0555 make sense for a FS mounted
+                     at a shared location?  */
+                  g_hash_table_insert (options, g_strdup("dmode"), g_strdup ("0555"));
+                }
+              else
+                {
+                  g_hash_table_insert (options, g_strndup (option, opt_len), g_strdup (value));
                 }
             }
           else
-            {
-              g_ptr_array_add (options, g_strdup (option));
-            }
+            g_hash_table_insert (options, g_strdup (option), NULL);
         }
     }
 
@@ -596,12 +591,24 @@ prepend_default_mount_options (const FSMountOptions *fsmo,
       gchar **split_option_string;
       split_option_string = g_strsplit (option_string, ",", -1);
       for (n = 0; split_option_string[n] != NULL; n++)
-        g_ptr_array_add (options, split_option_string[n]); /* steals string */
+        {
+          gchar *option = split_option_string[n];
+          const gchar *eq = strchr (option, '=');
+
+          if (eq != NULL)
+            {
+              const gchar *value = eq + 1;
+              gsize opt_len = eq - option;
+              g_hash_table_insert (options, g_strndup (option, opt_len), g_strdup (value));
+              g_free (option);
+            }
+          else
+            g_hash_table_insert (options, option, NULL); /* steals 'option' */
+        }
       g_free (split_option_string);
     }
-  g_ptr_array_add (options, NULL);
 
-  return (char **) g_ptr_array_free (options, FALSE);
+  return options;
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -706,58 +713,83 @@ calculate_mount_options (UDisksDaemon  *daemon,
                          GError       **error)
 {
   const FSMountOptions *fsmo;
-  gchar **options_to_use;
+  GHashTable *options_to_use = NULL;
+  GHashTableIter iter;
+  UDisksLinuxBlockObject *object = NULL;
+  UDisksLinuxDevice *device = NULL;
+  gboolean shared_fs = FALSE;
   gchar *options_to_use_str;
+  gchar *key, *value;
   GString *str;
-  guint n;
 
-  options_to_use = NULL;
   options_to_use_str = NULL;
 
   fsmo = find_mount_options_for_fs (fs_type);
 
+  object = udisks_daemon_util_dup_object (block, NULL);
+  device = udisks_linux_block_object_get_device (object);
+  if (device != NULL && device->udev_device != NULL &&
+      g_udev_device_get_property_as_boolean (device->udev_device, "UDISKS_FILESYSTEM_SHARED"))
+    shared_fs = TRUE;
+
+  g_clear_object (&device);
+  g_clear_object (&object);
+
   /* always prepend some reasonable default mount options; these are
    * chosen here; the user can override them if he wants to
    */
-  options_to_use = prepend_default_mount_options (fsmo, caller_uid, options);
+  options_to_use = prepend_default_mount_options (fsmo, caller_uid, options, shared_fs);
 
   /* validate mount options */
   str = g_string_new ("uhelper=udisks2,nodev,nosuid");
-  for (n = 0; options_to_use[n] != NULL; n++)
+  g_hash_table_iter_init (&iter, options_to_use);
+  while (g_hash_table_iter_next (&iter, (gpointer*) &key, (gpointer*) &value))
     {
-      const gchar *option = options_to_use[n];
-
       /* avoid attacks like passing "shortname=lower,uid=0" as a single mount option */
-      if (strstr (option, ",") != NULL)
+      if (strstr (key, ",") != NULL)
         {
           g_set_error (error,
                        UDISKS_ERROR,
                        UDISKS_ERROR_OPTION_NOT_PERMITTED,
                        "Malformed mount option `%s'",
-                       option);
+                       key);
           g_string_free (str, TRUE);
           goto out;
         }
 
       /* first check if the mount option is allowed */
-      if (!is_mount_option_allowed (fsmo, option, caller_uid))
+      if (!is_mount_option_allowed (fsmo, key, value, caller_uid))
         {
-          g_set_error (error,
-                       UDISKS_ERROR,
-                       UDISKS_ERROR_OPTION_NOT_PERMITTED,
-                       "Mount option `%s' is not allowed",
-                       option);
+          if (value == NULL)
+            {
+              g_set_error (error,
+                           UDISKS_ERROR,
+                           UDISKS_ERROR_OPTION_NOT_PERMITTED,
+                           "Mount option `%s' is not allowed",
+                           key);
+            }
+          else
+            {
+              g_set_error (error,
+                           UDISKS_ERROR,
+                           UDISKS_ERROR_OPTION_NOT_PERMITTED,
+                           "Mount option `%s=%s' is not allowed",
+                           key, value);
+            }
           g_string_free (str, TRUE);
           goto out;
         }
 
       g_string_append_c (str, ',');
-      g_string_append (str, option);
+      if (value == NULL)
+        g_string_append (str, key);
+      else
+        g_string_append_printf (str, "%s=%s", key, value);
     }
   options_to_use_str = g_string_free (str, FALSE);
 
  out:
-  g_strfreev (options_to_use);
+  g_hash_table_destroy (options_to_use);
 
   g_assert (options_to_use_str == NULL || g_utf8_validate (options_to_use_str, -1, NULL));
 
